#' @name defaults 
#' @aliases redDimPlotDefaults
#' @aliases featAssayPlotDefaults
#' @aliases colDataPlotDefaults 
#' @aliases rowStatTableDefaults
#' @aliases rowDataPlotDefaults
#' @aliases heatMapPlotDefaults
#'
#' @title Parameter defaults 
#'
#' @description Create default settings for various panels in the iSEE interface.
#'
#' @param se A SummarizedExperiment object.
#' @param number An integer scalar, specifying the maximum number of 
#' panels of the corresponding type that can be added to the interface.
#'
#' @section Reduced dimension plot parameters:
#' \describe{
#' \item{\code{Type}:}{Integer, which entry of \code{reducedDims(se)} should be shown?
#' Defaults to 1, i.e., the first entry.
#' Alternatively, a string can be supplied containing the name of the reduced dimension field, if \code{reducedDims(se)} has names.}
#' \item{\code{XAxis}:}{Integer, which component should be shown on the x-axis?
#' Defaults to 1.}
#' \item{\code{YAxis}:}{Integer, which component should be shown on the y-axis?
#' Defaults to 2.}
#' }
#'
#' @section Feature assay plot parameters:
#' \describe{
#' \item{\code{YAxisFeatName}:}{Integer, the index of the feature for which to show the expression on the y-axis if \code{YAxis="Feature name"}. 
#' Defaults to 1, i.e., the first feature in \code{se}.
#' Alternatively, a string can be supplied containing the name of the feature, i.e., the row name.}
#' \item{\code{YAxisRowTable}:}{Character, what row statistic table should be used to choose a feature to display on the y-axis?
#' Any setting will override \code{YAxisFeatName} upon initialization of the app.
#' Defaults to \code{"---"}, which means that no table will be used.}
#' \item{\code{Assay}:}{Integer, which assay should be used to supply the expression values shown on the y-axis?
#' Defaults to 1, i.e., the first assay in \code{se}.
#' Alternatively, a string can also be supplied containing the name of the assay, if \code{assays(se)} has names.}
#' \item{\code{XAxis}:}{Character, what type of variable should be shown on the x-axis?
#' Defaults to \code{"None"}, but can also be \code{"Row table"}, \code{"Column data"} or \code{"Feature name"}.}
#' \item{\code{XAxisColData}:}{Character, what column of \code{colData(se)} should be shown on the x-axis if \code{XAxis="Column data"}?
#' Defaults to the first entry of \code{colData(se)}.}
#' \item{\code{XAxisFeatName}:}{Integer, the index of the feature for which to show the expression on the x-axis if \code{XAxis="Feature name"}. 
#' Defaults to 1, i.e., the first feature in \code{se}.
#' Alternatively, a string can be supplied containing the name of the feature.}
#' \item{\code{XAxisRowTable}:}{Character, which row statistic table should be used to choose a feature to put on the x-axis if \code{XAxis="Row table"}? 
#' Any setting will override \code{XAxisFeatName} upon initialization of the app.
#' Defaults to \code{"---"}, which means that no table will be used.}
#' }
#'
#' @section Column data plot parameters:
#' \describe{
#' \item{\code{YAxis}:}{Character, which column of \code{colData(se)} should be shown on the y-axis?
#' Defaults to the first entry of \code{colData(se)}.}
#' \item{\code{XAxis}:}{Character, what type of variable should be shown on the x-axis?
#' Defaults to \code{"None"}, but can also be \code{"Column data"}.}
#' \item{\code{XAxisColData}:}{Character, which column of \code{colData(se)} should be shown on the x-axis if \code{XAxis="Column data"}?
#' Defaults to the first entry of \code{colData(se)}.}
#' }
#'
#' @section Row data plot parameters:
#' \describe{
#' \item{\code{YAxis}:}{Character, which column of \code{rowData(se)} should be shown on the y-axis?
#' Defaults to the first entry of \code{colData(se)}.}
#' \item{\code{XAxis}:}{Character, what variable should be shown on the x-axis?
#' Defaults to \code{"None"}, but can also be \code{"Row table"}, \code{"Row data"} or \code{"Feature name"}.}
#' \item{\code{XAxisRowData}:}{Character, which column of \code{rowData(se)} should be shown on the x-axis if \code{XAxis="Row data"}?
#' Defaults to the first entry of \code{rowData(se)}.}
#' }
#'
#' @section Coloring parameters (for points):
#' \describe{
#' \item{\code{ColorBy}:}{Character, what type of data should be used for coloring?
#' Defaults to \code{"None"}, but can also be \code{"Row table"}, \code{"Feature name"} or \code{"Column data"}.}
#' \item{\code{ColorByDefaultColor}:}{String specifying the default point colour when \code{ColorBy="None"}.
#' Defaults to \code{"black"}.}
#' \item{\code{ColorByFeatName}:}{Integer, the index of the feature to use for colouring based on expression, if \code{ColorBy="Feature name"}? 
#' Defaults to 1, i.e., the first feature in \code{se}.
#' Alternatively, a string can be supplied containing the name of the feature.}
#' \item{\code{ColorByRowTable}:}{Character, which row statistic table should be used to choose a feature to color by, if \code{ColorBy="Feature name"}? 
#' Any setting will override \code{ColorByFeatName} upon initialization of the app.
#' Defaults to \code{"---"}, which means that no table will be used.}
#' }
#'
#' For the plots where each point represents a sample (i.e., all plots except for heatmaps and row data plots), the following additional options apply:
#' \describe{
#' \item{\code{ColorByColData}:}{Character, which column of \code{colData(se)} should be used for colouring if \code{ColorBy="Column data"}? 
#' Defaults to the first entry of \code{colData(se)}.}
#' \item{\code{ColorByFeatNameAssay}:}{Integer, which assay should be used to supply the expression values for colouring if \code{ColorBy="Feature name"}? 
#' Defaults to 1, i.e., the first assay in \code{se}.
#' Alternatively, a string can also be supplied containing the name of the assay, if \code{assays(se)} has names.}
#' }
#'
#' For plots where each point represents a feature (i.e., row data plots), the following additional options apply:
#' \describe{
#' \item{\code{ColorByRowData}:}{Character, which column of \code{rowData(se)} should be used for colouring if \code{ColorBy="Row data"}? 
#' Defaults to the first entry of \code{rowData(se)}.}
#' \item{\code{ColorByFeatNameColor}:}{String specifying the colour to be used to highlight the selected feature from the text if \code{ColorBy="Feature name"}.
#' Defaults to \code{"red"}.}
#' }
#'
#' @section Other plot parameters (for point-based plots):
#' \describe{
#' \item{\code{DataBoxOpen}:}{Logical, should the data parameter box be open upon initialization?
#' Defaults to \code{FALSE}.}
#' \item{\code{VisualBoxOpen}:}{Logical, should the visual parameter box be open upon initialization?
#' Defaults to \code{FALSE}.}
#' \item{\code{VisualChoices}:}{A list containing one character vector, specifying the visual box parameters to be shown upon initialization.
#' This defaults to \code{list("Color")} to show only the coloring parameters, but the internal vector can also contain \code{"Points"} and \code{"Other"}.}
#' \item{\code{PointSize}:}{Numeric, the size of the points.
#' Defaults to 1.}
#' \item{\code{PointAlpha}:}{Numeric, what level of transparency should be used for the points?
#' Ignored when \code{SelectEffect="Transparent"} and the transmitting plot has a non-\code{NULL} selection of points.
#' Defaults to 1.}
#' \item{\code{Downsample}:}{Logical, indicating whether downsampling of overlapping points should be performed.
#' Defaults to \code{FALSE}.}
#' \item{\code{SampleRes}:}{Numeric, specifying the downsampling resolution, i.e., the granularity at which points are considered to overlap.
#' Higher values result in a more stringent definition of overlaps, and thus less downsampling.
#' Defaults to 200.}
#' \item{\code{FontSize}:}{Numeric, size of the font.
#' Defaults to 1.}
#' \item{\code{LegendPosition}:}{String specifying the legend position.
#' Defaults to \code{"Bottom"} but can also be \code{"Right"}.}
#' \item{\code{ZoomData}:}{A list containing numeric vectors of length 4, containing values with names \code{"xmin"}, \code{"xmax"}, \code{"ymin"} and \code{"ymax"}.
#' These define the zoom window on the x- and y-axes. 
#' Each element of the list defaults to \code{NULL}, i.e., no zooming is performed.}
#' }
#' 
#' @section Row statistics table parameters:
#' \describe{
#' \item{\code{Selected}:}{Integer, containing the index of the row to be initially selected.
#' Defaults to the first row, i.e., 1.
#' Alternatively, a string can be supplied containing the row name.}
#' \item{\code{Search}:}{Character, containing the initial value of the search field.
#' Defaults to an empty string.}
#' \item{\code{SearchColumns}:}{A list containing character vectors of length equal to the number of columns in \code{rowData(se)},
#' specifying the initial value of the search field for each column.
#' All entries default to an empty string.}
#' }
#'
#' @section Heatmap parameters:
#' The features/rows to be used in the construction of the heatmap are specified with:
#' \describe{
#' \item{\code{FeatName}:}{List of length equal to the number of panels.
#' Each list entry corresponds to a panel and should be an integer vector with the indices of the feature(s) for which to show the expression in the heatmap. 
#' Defaults to \code{1L} for each panel, i.e., the first feature in \code{se}.
#' Alternatively, a character vector can be supplied containing the names of the features.}
#' \item{\code{Assay}:}{Integer, which assay should be used to supply the expression values shown on the y-axis?
#' Defaults to 1, i.e., the first assay in \code{se}.
#' Alternatively, a string can also be supplied containing the name of the assay, if \code{assays(se)} has names.}
#' \item{\code{FeatNameSource}:}{Character, which other panel should be used to choose the features to show in the heatmap?
#'  Defaults to \code{"---"}, which means that no panel is used for feature selection.}
#' \item{\code{FeatNameBoxOpen}:}{Logical, should the feature selection box be open upon initialization?
#' Defaults to \code{FALSE}.}
#' }
#'
#' The column metadata variables control the ordering of the samples in the heatmap. 
#' They can be controlled with:
#' \describe{
#' \item{\code{ColData}:}{List of length equal to the number of panels.
#' Each list entry corresponds to a panel and should contain a character vector specifying the field(s) of \code{colData(se)} that should be used to order the samples in the heatmap.
#' Note that these fields will also appear as annotation bars.
#' Each character vector defaults to the first entry of \code{colData(se)}.}
#' \item{\code{ColDataBoxOpen}:}{Logical, should the column data selection box be open upon initialization?
#' Defaults to \code{FALSE}.}
#' }
#'
#' A variety of parameters are available to control the color scale of the heatmap.
#' They can be specified with:
#' \describe{ 
#' \item{\code{ColorBoxOpen}:}{Logical, should the color selection panel for the heatmap be open upon initialization?
#' Defaults to \code{FALSE}.}
#' \item{\code{CenterScale}:}{List of length equal to the number of panels.
#' Each list entry corresponds to a panel and contains a character vector specifying whether each row of expression values should be mean-centered and/or scaled to unit variance.
#' Defaults to \code{"Centered"}. 
#' Users can set it to \code{c("Centered", "Scaled")} to obtain mean-centered and unit-scaled rows.}
#' \item{\code{Lower}:}{Numeric, what should be the lower bound of the color scale for the values in the heatmap? All values below this threshold will be shown in the same color.
#' Defaults to -Inf, meaning that the lowest value in the data matrix will be used.}
#' \item{\code{Upper}:}{Numeric, what should be the upper bound of the color scale for the values in the heatmap? All values above this threshold will be shown in the same color.
#' Defaults to Inf, meaning that the highest value in the data matrix will be used.}
#' \item{\code{ColorScale}:}{Character, what color scale (in the form low-mid-high) should be used to color the heatmap when values are centered?
#' Defaults to \code{"purple-black-yellow"}.}
#' }
#'
#' Finally, the \code{ZoomData} field for heatmaps should contain an integer vector of consecutive indices to zoom into from the full heatmap. 
#' This vector will subset the entries in \code{FeatName} for a given panel. 
#' This defaults to \code{NULL}, i.e., all specified features in \code{FeatName} are shown.
#'
#' @section Point selection parameters:
#' For the point-based plots, the following options apply:
#' \describe{
#' \item{\code{SelectBoxOpen}:}{Logical, should the point selection parameter box be open upon initialization?
#' Defaults to \code{FALSE}.}
#' \item{\code{SelectByPlot}:}{Character, which other plot should be used for point selection in the current plot? 
#' Defaults to \code{"---"}, which means that no plot is used for point selection.}
#' \item{\code{SelectEffect}:}{Character, what is the effect of receiving point selection information?
#' Can be \code{"Restrict"}, where only the selected points are shown; \code{"Color"}, where the selected points have a different color; 
#' or \code{"Transparent"}, where all points other than those selected are made transparent. 
#' Defaults to \code{"Transparent"}.}
#' \item{\code{SelectColor}:}{Character, what color should be used for the selected points when \code{SelectEffect="Color"}?
#' Defaults to \code{"red"}.}
#' \item{\code{SelectAlpha}:}{Numeric, what level of transparency should be used for the unselected points when \code{SelectEffect="Transparent"}?
#' This should lie in [0, 1], where 0 is fully transparent and 1 is fully opaque. 
#' Defaults to 0.1.}
#' }
#'
#' Note that point selection cannot occur between row data plots and the other plot types.
#' This is because each point in a row data plot is a feature, while each point represents a sample in the other plots.
#' Point selection can only occur between plots of the same point type.
#'
#' For the row statistics tables, the following options apply:
#' \describe{
#' \item{\code{SelectBoxOpen}:}{Logical, should the point selection parameter box be open upon initialization?
#' Defaults to \code{FALSE}.}
#' \item{\code{SelectByPlot}:}{Character, which other plot should be used to select features in the current table? 
#' Defaults to \code{"---"}, which means that no plot is used for point selection.}
#' }
#' Only row data plots can be used for selecting points to supply to tables.
#' 
#' @return A DataFrame containing default settings for various parameters of each panel.
#'
#' @export
#'
#' @examples
#' library(scRNAseq)
#' data(allen)
#' class(allen)
#'
#' library(scater)
#' sce <- as(allen, "SingleCellExperiment")
#' counts(sce) <- assay(sce, "tophat_counts")
#' sce <- normalize(sce)
#' sce <- runPCA(sce)
#' sce
#'
#' redDimPlotDefaults(sce, number=5)
#' featAssayPlotDefaults(sce, number=5)
#' colDataPlotDefaults(sce, number=5)
#' rowStatTableDefaults(sce, number=5)
#' rowDataPlotDefaults(sce, number=5)
#' heatMapPlotDefaults(sce, number=5)
redDimPlotDefaults <- function(se, number) {
    waszero <- number==0 # To ensure that we define all the fields with the right types.
    if (waszero) number <- 1

    out <- new("DataFrame", nrows=as.integer(number))
    out[[.redDimType]] <- 1L
    out[[.redDimXAxis]] <- 1L
    out[[.redDimYAxis]] <- 2L
    
    out <- .add_general_parameters_for_column_plots(out, se)
    if (waszero) out <- out[0,,drop=FALSE]
    return(out)
}

#' @rdname defaults 
#' @export
featAssayPlotDefaults <- function(se, number) {
    waszero <- number==0 
    if (waszero) number <- 1

    def_assay <- .set_default_assay(se)
    covariates <- colnames(colData(se))

    out <- new("DataFrame", nrows=as.integer(number))
    out[[.featAssayAssay]] <- def_assay
    out[[.featAssayXAxis]] <- .featAssayXAxisNothingTitle
    out[[.featAssayXAxisColData]] <- covariates[1] 
    out[[.featAssayXAxisFeatName]] <- 1L
    out[[.featAssayXAxisRowTable]] <- .noSelection
    out[[.featAssayYAxisFeatName]] <- 1L
    out[[.featAssayYAxisRowTable]] <- .noSelection

    out <- .add_general_parameters_for_column_plots(out, se)
    if (waszero) out <- out[0,,drop=FALSE]
    return(out)
}

#' @rdname defaults 
#' @export
colDataPlotDefaults <- function(se, number) {
    waszero <- number==0 
    if (waszero) number <- 1

    covariates <- colnames(colData(se))

    out <- new("DataFrame", nrows=as.integer(number))
    out[[.colDataYAxis]] <- covariates[1]
    out[[.colDataXAxis]] <- .colDataXAxisNothingTitle
    out[[.colDataXAxisColData]] <- ifelse(length(covariates)==1L, covariates[1], covariates[2])

    out <- .add_general_parameters_for_column_plots(out, se)
    if (waszero) out <- out[0,,drop=FALSE]
    return(out)
}

#' @rdname defaults
#' @export
rowStatTableDefaults <- function(se, number) {
    waszero <- number==0 
    if (waszero) number <- 1

    out <- new("DataFrame", nrows=as.integer(number))
    out[[.rowStatSelected]] <- 1L
    out[[.rowStatSearch]] <- ""

    # Defining an empty search for each column of the rowData.
    colsearch <- character(ncol(rowData(se)))
    out[[.rowStatColSearch]] <- rep(list(colsearch), as.integer(number))

    # Defining the rowDataPlot from which point selections are received.
    out[[.selectParamBoxOpen]] <- FALSE
    out[[.selectByPlot]] <- .noSelection

    if (waszero) out <- out[0,,drop=FALSE]
    return(out)
}

#' @rdname defaults 
#' @export
rowDataPlotDefaults <- function(se, number) {
    waszero <- number==0 
    if (waszero) number <- 1

    covariates <- colnames(rowData(se))

    out <- new("DataFrame", nrows=as.integer(number))
    out[[.rowDataYAxis]] <- covariates[1]
    out[[.rowDataXAxis]] <- .rowDataXAxisNothingTitle
    out[[.rowDataXAxisRowData]] <- ifelse(length(covariates)==1L, covariates[1], covariates[2])

    out <- .add_general_parameters_for_row_plots(out, se)
    if (waszero) out <- out[0,,drop=FALSE]
    return(out)
}

#' @rdname defaults 
#' @export
heatMapPlotDefaults <- function(se, number) {
    waszero <- number==0 # To ensure that we define all the fields with the right types.
    if (waszero) number <- 1
    
    def_assay <- .set_default_assay(se)
    
    out <- new("DataFrame", nrows=as.integer(number))
    out[[.heatMapAssay]] <- def_assay
    out[[.heatMapFeatNameBoxOpen]] <- FALSE
    out[[.heatMapFeatName]] <- rep(list(1L), nrow(out))
  
    out[[.heatMapColDataBoxOpen]] <- FALSE
    out[[.heatMapColData]] <- rep(list(colnames(colData(se))[1]), nrow(out))
    out[[.heatMapImportSource]] <- .noSelection
  
    out[[.heatMapCenterScale]] <- rep(list(.heatMapCenterTitle), nrow(out))
    out[[.heatMapLower]] <- -Inf
    out[[.heatMapUpper]] <- Inf
    out[[.heatMapCenteredColors]] <- "purple-black-yellow"
    
    out[[.zoomData]] <- rep(list(NULL), nrow(out))

    out[[.selectParamBoxOpen]] <- FALSE
    out[[.selectByPlot]] <- .noSelection
    out[[.selectEffect]] <- .selectTransTitle
    out[[.selectTransAlpha]] <- 0.1
    out[[.selectColor]] <- "red"
  
    if (waszero) out <- out[0,,drop=FALSE]
    return(out)
}

#' Override default parameters
#'
#' Override the default settings of various parameters with whatever the user has supplied.
#'
#' @param def A DataFrame of default values, generated using \code{\link{redDimPlotDefaults}} or similar functions.
#' @param usr A DataFrame or data.frame of user-specified values, to use to replace the defaults.
#' 
#' @return A DataFrame with the default parameter settings replaced by user-specified values, where appropriate.
#'
#' @details
#' Not all arguments in \code{def} need to be specified in \code{usr}.
#' Parameters will only be overridden for the specified arguments.
#'
#' This function expects that \code{nrow(def)} is greater than or equal to \code{nrow(usr)}.
#' Parameters in \code{def} will only be overwritten for the first \code{nrow(usr)} panels.
#' 
#' @author Aaron Lun
#' @rdname INTERNAL_override_defaults
#' @seealso
#' \code{?\link{defaults}},
#' \code{\link{.setup_memory}}
.override_defaults <- function(def, usr)
{
    ndef <- nrow(def)
    nusr <- nrow(usr)
    stopifnot(ndef >= nusr) 
    replacement <- seq_len(nusr)

    for (x in colnames(usr)) {
        if (!x %in% colnames(def)) { 
            warning(sprintf("unknown field '%s' in user-specified settings", x))
            next 
        }

        # This method is safer than direct subset assignment, 
        # as it works properly for lists.
        tmp <- def[[x]]
        tmp[replacement] <- usr[[x]]
        def[[x]] <- tmp 
    }

    return(def)
}    

#' Add general plot defaults 
#'
#' Add default values for general plot parameters, including row and column-based plots.
#'
#' @param incoming A DataFrame with non-zero number of rows, containing default parameters that have already been filled for specific panel types.
#' 
#' @return A DataFrame with additional fields for general plot parameters, filled with default values.
#' 
#' @details
#' The \code{.add_general_parameters} function adds general parameters such as parameter box opening flags, 
#' point selection specifications, and zoom, brush and lasso data fields.
#' All parameters are initialized at their default values.
#'
#' The \code{.add_general_parameters_for_column_plots} function adds general parameters for column-based plots,
#' while the \code{.add_general_parameters_for_row_plots} function adds them for row-based plots.
#' These mainly differ in how colouring is performed.
#' 
#' @author Aaron Lun
#' @rdname INTERNAL_add_general_parameters
#' @seealso
#' \code{\link{defaults}}
.add_general_parameters <- function(incoming) {
    incoming[[.dataParamBoxOpen]] <- FALSE
    incoming[[.visualParamBoxOpen]] <- FALSE
    incoming[[.selectParamBoxOpen]] <- FALSE

    incoming[[.selectByPlot]] <- .noSelection
    incoming[[.selectEffect]] <- .selectTransTitle
    incoming[[.selectTransAlpha]] <- 0.1
    incoming[[.selectColor]] <- "red"
    incoming[[.brushData]] <- rep(list(NULL), nrow(incoming))

    incoming[[.visualParamChoice]] <- rep(list(.visualParamChoiceColorTitle), nrow(incoming))
   
    incoming[[.plotPointSize]] <- 1 
    incoming[[.plotPointAlpha]] <- 1
    incoming[[.plotPointDownsample]] <- FALSE
    incoming[[.plotPointSampleRes]] <- 200
    
    incoming[[.plotFontSize]] <- 1
    incoming[[.plotLegendPosition]] <- .plotLegendBottomTitle

    incoming[[.zoomData]] <- rep(list(NULL), nrow(incoming))
    incoming[[.lassoData]] <- rep(list(NULL), nrow(incoming))
    return(incoming)
}

#' @param se A SummarizedExperiment object.
#' @rdname INTERNAL_add_general_parameters
.add_general_parameters_for_column_plots <- function(incoming, se) {
    incoming <- .add_general_parameters(incoming)

    # Adding coloring parameters specifically for column plots.
    def_assay <- .set_default_assay(se)
    def_cov <- colnames(colData(se))[1]
    incoming[[.colorByField]] <- .colorByNothingTitle
    incoming[[.colorByDefaultColor]] <- "black"
    incoming[[.colorByColData]] <- def_cov
    incoming[[.colorByRowTable]] <- .noSelection 
    incoming[[.colorByFeatName]] <- 1L
    incoming[[.colorByFeatNameAssay]] <- def_assay

    return(incoming)
}

#' @rdname INTERNAL_add_general_parameters
.add_general_parameters_for_row_plots <- function(incoming, se) {
    incoming <- .add_general_parameters(incoming)

    # Adding coloring parameters specifically for row plots.
    def_cov <- colnames(rowData(se))[1]
    incoming[[.colorByField]] <- .colorByNothingTitle
    incoming[[.colorByDefaultColor]] <- "black"
    incoming[[.colorByRowData]] <- def_cov
    incoming[[.colorByRowTable]] <- .noSelection 
    incoming[[.colorByFeatName]] <- 1L
    incoming[[.colorByFeatNameColor]] <- "red"

    return(incoming)
}

#' Set the default assay
#'
#' Identifies the index of the assay containing the log-count matrix.
#'
#' @param se A SummarizedExperiment object.
#' 
#' @return An integer scalar containing the index of the \code{"logcounts"} assay, if available; otherwise 1L.
#'
#' @details
#' It usually makes most sense to perform visualization (colouring, examination of expression) on the log-expression values,
#' as this provides good dynamic range and easy interpretability.
#' This function conveniently identifies the index of the assay named \code{"logcounts"} for use as the default when it is available.
#'
#' @author Aaron Lun
#' @rdname INTERNAL_set_default_assay
#' @seealso
#' \code{?\link{defaults}}
.set_default_assay <- function(se) { 
    def_assay <- which(assayNames(se)=="logcounts")
    if (length(def_assay)==0L) {
        return(1L)
    } else {
        return(def_assay[1])
    }
}
